/*-------------------------------------------------------------------------*/
/* xcircuit.h 								   */
/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/*      written by Tim Edwards, 8/20/93    				   */
/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/* Definition shortcuts 						   */
/*-------------------------------------------------------------------------*/

#define XtnSetArg(a,b)  	XtSetArg(wargs[n], a, b); n++
#define abs(a)			((a) < 0 ? -(a) : (a))
#define sign(a)			((a) <= 0 ? -1 : 1)
#define max(a,b)		((a) < (b) ? (b) : (a))
#define min(a,b)		((a) < (b) ? (a) : (b))
#define Number(a)		(void *)((unsigned int) a)
#define ROT_UNIT (360 / RSTEPS) /* degrees per xcircuit rotational unit */

/*-------------------------------------------------------------------------*/
/* Lengthier define constructs						   */
/*-------------------------------------------------------------------------*/

#define objectdata      areastruct.topobject->thisobject
#define PLIST_INCR(a) \
	a->plist = (genericptr *) realloc (a->plist, \
	(a->parts + 1) * sizeof(genericptr))
#define SELECTTYPE(a)   (*(objectdata->plist + *a))->type
#define SELTOCOLOR(a)	(*(objectdata->plist + *a))->color
#define GSELTOCOLOR(b, a) (*(b->plist + *a))->color
#define ENDPART		objectdata->plist + objectdata->parts
#define EDITPART	objectdata->plist + areastruct.editpart

/* conversions from generic to specific types */
/* specifically, from type (generic *) to type (polyptr), etc. */

#define TOPOLY(a)	(*((polyptr *)a))
#define TOLABEL(a)	(*((labelptr *)a))
#define TOOBJINST(a)	(*((objinstptr *)a))
#define TOARC(a)	(*((arcptr *)a))
#define TOSPLINE(a)	(*((splineptr *)a)) 
#define TOPATH(a)	(*((pathptr *)a))

/* conversions from a selection to a specific type */

#define SELTOPOLY(a)	TOPOLY(objectdata->plist + (*a))
#define SELTOLABEL(a)	TOLABEL(objectdata->plist + (*a))
#define SELTOOBJINST(a)	TOOBJINST(objectdata->plist + (*a))
#define SELTOARC(a)	TOARC(objectdata->plist + (*a))
#define SELTOSPLINE(a)	TOSPLINE(objectdata->plist + (*a))
#define SELTOPATH(a)	TOPATH(objectdata->plist + (*a))

/* creation of new elements */

#define NEW_POLY(a,b) \
	PLIST_INCR(b); \
	a = (polyptr *)b->plist + b->parts; \
	*a = (polyptr) malloc(sizeof(polygon)); \
	(*a)->type = POLYGON
#define NEW_LABEL(a,b) \
	PLIST_INCR(b); \
	a = (labelptr *)b->plist + b->parts; \
	*a = (labelptr) malloc(sizeof(label)); \
	(*a)->type = LABEL
#define NEW_OBJINST(a,b) \
	PLIST_INCR(b); \
	a = (objinstptr *)b->plist + b->parts; \
	*a = (objinstptr) malloc(sizeof(objinst)); \
	(*a)->type = OBJECT
#define NEW_ARC(a,b) \
	PLIST_INCR(b); \
	a = (arcptr *)b->plist + b->parts; \
	*a = (arcptr) malloc(sizeof(arc)); \
	(*a)->type = ARC
#define NEW_SPLINE(a,b) \
	PLIST_INCR(b); \
	a = (splineptr *)b->plist + b->parts; \
	*a = (splineptr) malloc(sizeof(spline)); \
	(*a)->type = SPLINE
#define NEW_PATH(a,b) \
	PLIST_INCR(b); \
	a = (pathptr *)b->plist + b->parts; \
	*a = (pathptr) malloc(sizeof(path)); \
	(*a)->type = PATH

/* selection from top-level object */

#define objectselect(a)		selectobject(a, objectdata)
#define easydraw(a, b)		geneasydraw(a, b, objectdata)
#define drawselected(a)		gendrawselected(a, objectdata)
#define calcbbox(a)		calcbboxvalues(a, (uchar)0)
		   
/*-------------------------------------------------------------------------*/

typedef unsigned char		uchar;

typedef struct {
   float x, y;
} XfPoint;

typedef struct {
   long x, y;
} XlPoint;

/*-------------------------------------------------------------------------*/
/* Implementation-specific definitions 					   */
/*-------------------------------------------------------------------------*/

#define LIBS	    3 /* number of library pages 			*/
#define PAGES      10 /* default number of top-level pages 		*/
#define SCALEFAC  1.5 /* zoom in/out scaling multiplier 		*/
#define SUBSCALE 0.67 /* ratio of subscript size to normal script size 	*/
#define SBARSIZE   13 /* Pixel size of the scrollbar 			*/
#define RSTEPS     72 /* Number of allowed rotation positions 		*/
#define SPLINESEGS 20 /* Number of points per spline approximation 	*/
#define DELBUFSIZE 10 /* Number of delete events to save for undeleting */
#define MINAUTOSCALE 0.75 /* Won't automatically scale closer than this */

#ifdef SCHEMA
#define PADSPACE   20 /* Spacing of pinlabels from their origins	*/
#endif

#define TOPLEVEL	0
#define SINGLE		1

#define INTSEGS		(SPLINESEGS - 2)
#define LASTSEG		(SPLINESEGS - 3)

#define FONTHEIGHT(y)	(y->ascent + y->descent + 6) 
#define ROWHEIGHT	FONTHEIGHT(appdata.xcfont)
#define FILECHARASCENT  (appdata.filefont->ascent)
#define FILECHARHEIGHT  (FILECHARASCENT + appdata.filefont->descent)
#define LISTHEIGHT	200
#define TEXTHEIGHT	28 /* Height of xcircuit vectored font at nominal size */
#define HALFHEIGHT	14 /* Half of TEXTHEIGHT */

/*-----------------------------------------------------------------------*/

#define INVRFAC 57.295779       /* 180 / pi */

/*-----------------------------------------------------------------------*/

#define INCHSCALE 0.375     /* Scale of .25 inches to PostScript units */
#define CMSCALE 0.35433071  /* Scale of .5 cm to PostScript units */
#define IN_CM_CONVERT 28.3464567	/* 72 (in) / 2.54 (cm/in) */

/*-------------------------------------------------------------------------*/
/* Event mode definitions (state of drawing area)			   */
/*-------------------------------------------------------------------------*/

enum editmode {NORMAL_MODE = 0, 
PRESS_MODE, DELETE_MODE, COPY_MODE, COPY2_MODE, ROTATE_MODE,
PUSH_MODE, EDIT_MODE, SELAREA_MODE, SELAREA2_MODE, DESEL_MODE, PAN_MODE,
CATPAN_MODE, CATALOG_MODE, CATTEXT_MODE, FONTCAT_MODE, FONTCAT2_MODE,
TEXT1_MODE, TEXT2_MODE, TEXT3_MODE, WIRE_MODE, BOX0_MODE, BOX_MODE,
EBOX_MODE, ARC0_MODE, ARC_MODE, EARC_MODE, SPLINE0_MODE, SPLINE_MODE,
ESPLINE_MODE, EPATH_MODE, PARAM_MODE,

#ifdef SCHEMA
CONNECT_MODE,
#endif

};

/*-------------------------------------------------------------------------*/
/* Text justification styles						   */
/*-------------------------------------------------------------------------*/

#define NOTLEFT		1
#define RIGHT		2
#define NOTBOTTOM	4
#define TOP		8
#define FLIPINV		16

/*-------------------------------------------------------------------------*/
/* Text script styles						   	   */
/*-------------------------------------------------------------------------*/

#define TEXT_ESC	27  /* escape character for defining text controls */
#define CONTROL_START	32  /* first character used by controls */
#define SUBSCRIPT	32
#define SUPERSCRIPT	33
#define NORMALSCRIPT	34
#define UNDERLINE	35
#define OVERLINE	36
#define NOLINE		37
#define BACKSPACE	38
#define HALFSPACE	39
#define QTRSPACE	40
#define SPECIAL		41
#define FONT_SCALE	42    /* inline font scale designator (begin) */
#define END_SCALE	43    /* inline font scale designator (end) */
#define PARAM_START	44
#define PARAM_END	45

#define FONT_START     100    /* character of first inline font designator */

#define NO_OP	       255    /* padding for parameterized substrings */


/*-------------------------------------------------------------------------*/
/* Reset modes								   */
/*-------------------------------------------------------------------------*/

#define SAVE		1
#define DESTROY		2

/*-------------------------------------------------------------------------*/
/* Coordinate display types						   */
/*-------------------------------------------------------------------------*/

#define DEC_INCH	0
#define FRAC_INCH	1
#define CM		2

/*-------------------------------------------------------------------------*/
/* Library types							   */
/*-------------------------------------------------------------------------*/

#define USERLIB		0
#define LIBRARY		1
#define FONTLIB		2

/*-------------------------------------------------------------------------*/
/* Box styles								   */
/*-------------------------------------------------------------------------*/

#define NORMAL 0
#define UNCLOSED 1
#define DASHED 2
#define DOTTED 4
#define NOBORDER 8
#define FILLED 16
#define STIP0 32
#define STIP1 64
#define STIP2 128
#define FILLSOLID 224  /* = 32 + 64 + 128 */
#define OPAQUE 256

/*-------------------------------------------------------------------------*/
/* Box edit styles							   */
/*-------------------------------------------------------------------------*/

#define MANHATTAN 1
#define RHOMBOIDX 2
#define RHOMBOIDY 4
#define RHOMBOIDA 8

#define NONE  0
#define LASTX 1
#define NEXTX 2
#define LASTY 4
#define NEXTY 8

/*-------------------------------------------------------------------------*/
/* Arc creation and edit styles						   */
/*-------------------------------------------------------------------------*/

#define CENTER 1
#define RADIAL 2

/*-------------------------------------------------------------------------*/
/* Delete/undelete draw-mode styles					   */
/*-------------------------------------------------------------------------*/

#define ERASE 1
#define DRAW  1

/*-------------------------------------------------------------------------*/
/* Schematic object types and pin label types	   			   */
/*-------------------------------------------------------------------------*/

#ifdef SCHEMA

#define SCHEMATIC 0
#define FUNDAMENTAL 1
#define SYMBOL 2

#define LOCAL 1
#define GLOBAL 2
#define INFO 3

#endif

/*-------------------------------------------------------------------------*/
/* Color scheme styles (other than NORMAL)				   */
/*-------------------------------------------------------------------------*/

#define INVERSE		1

/*-------------------------------------------------------------------------*/
/* Cursor definitions 							   */
/*-------------------------------------------------------------------------*/

#define NUM_CURSORS	9

#define ARROW 		appcursors[0]
#define CROSS 		appcursors[1]
#define SCISSORS 	appcursors[2]
#define COPYCURSOR	appcursors[3]
#define ROTATECURSOR    appcursors[4]
#define EDCURSOR	appcursors[5]
#define TEXTPTR 	appcursors[6]
#define CIRCLE		appcursors[7]
#define QUESTION	appcursors[8]

/*-------------------------------------------------------------------------*/
/* selection-mechanism structures					   */
/*-------------------------------------------------------------------------*/

typedef XPoint* pointlist;
typedef XfPoint* fpointlist;

/*-------------------------------------------------------------------------*/
/* select types */
/*-------------------------------------------------------------------------*/

#define	OBJECT	1
#define LABEL	2
#define POLYGON	4
#define ARC	8
#define SPLINE  16
#define PATH	32
#define ANY	63  /* for selection mechanism only */

/*--------------------------------------------------------------------------*/
/* structures of all the main elements which can be displayed & manipulated */
/*--------------------------------------------------------------------------*/

typedef struct _object *objectptr;

typedef struct {
   uchar	type;		/* type is LABEL, POLYGON, etc., from below */
   int		color;
} generic, *genericptr;		/* (convenience function for retypecasting) */

typedef struct {
   uchar	type;
   int		color;
   XPoint	position;
   short	rotation;
   float	scale;
   void		**params;	/* parameter types determined from thisobject */
   uchar	num_params;	/* number of parameters currently allocated */
   objectptr	thisobject;
} objinst, *objinstptr;

typedef struct {
   objectptr    thisobject;
   objinstptr	thisinst;
} objectpair;

typedef struct {
   uchar	type;
   int		color;
   XPoint	position;
   short	rotation;
   float	scale;
   short	justify;
#ifdef SCHEMA
   uchar	pin;
#endif
   uchar	*string;
} label, *labelptr;

typedef struct {
   uchar	type;
   int		color;
   short	style;
   float	width;
   short	number;
   pointlist	points;
} polygon, *polyptr;

typedef struct {
   uchar	type;
   int		color;
   short	style;
   float	width;
   XPoint	ctrl[4];
   /* the following are for rendering only */
   XfPoint	points[INTSEGS];
} spline, *splineptr;

typedef struct {
   uchar	type;
   int		color;
   short	style;
   float	width;
   short	radius; 	/* x-axis radius */
   short	yaxis;		/* y-axis radius */
   float	angle1;
   float	angle2;
   XPoint	position;
   /* the following are for rendering only */
   short	number;
   XfPoint	points[RSTEPS + 1];
} arc, *arcptr;

typedef struct {
   uchar	type;
   int		color;
   short	style;
   float	width;
   short	parts;
   genericptr	*plist;		/* to be retypecast to polygon, arc, or spline */
} path, *pathptr;

/*--------------------------------------------------------------------------*/
/* Netlist structures for schematic capture				    */
/* (courtesy of Chow Seong Hwai, 3/20/98)				    */
/*--------------------------------------------------------------------------*/

#ifdef SCHEMA

struct Netlist
{
   polyptr poly;
   int netid;
   struct Netlist *next;
};

struct Pinlist
{
   labelptr pin;
   char *pinx;
   int netid;
   struct Pinlist *next;
};

struct Paramlist
{
   int netid;
   labelptr pin;
   struct Paramlist *next;
};

struct Calllist
{
   objectptr callobj;
   objinstptr callno;
   struct Paramlist *params;
   struct Calllist *next;
};

/* PCB netlist structures */

struct Pnet {
   int netidx;
   struct Pnet *next;
};

struct Pstr {
   char *string;
   struct Pstr *next;
};

struct Ptab {
   struct Pnet *nets;
   struct Pstr *pins;
   struct Ptab *next;
};

#endif

/*----------------------------------------------------------------------*/
/* Object parameter structure						*/
/*----------------------------------------------------------------------*/

enum paramtypes {XC_SHORT = 0, XC_STRING};

typedef struct _oparam {
   uchar	type;		/* type is from paramtypes list above */
   void		*pdefault;	/* default value */
   uchar	number;		/* number of locations to replace */
   void		**position;	/* list of locations to replace (fragile!) */
} oparam, *oparamptr;

/*--------------------------------------------------------------------------*/
/* Main object structure						    */
/*--------------------------------------------------------------------------*/

typedef struct _object {
   char		name[80];
   Boolean	hidden;
   XPoint	lowerleft;	/* bounding box */
   Dimension	width, height;  /* information  */
   short	parts;
   genericptr	*plist;	 	/* to be retypecast to label, polygon, etc. */
   uchar	num_params;	/* number of calling parameters */
   oparamptr	*params;	/* list of parameter values */
   
#ifdef SCHEMA
   XPoint	lleft2;		 /* bounding box info excluding pins */
   Dimension	width2, height2;

   uchar	schemtype;
   objectptr	symschem;	 /* schematic page support */

   struct Netlist   *netlist;
   struct Pinlist   *pinlist;
   struct Calllist  *calllist;
   short traversed;
#endif
} object;

/*--------------------------------------------------------------------------*/
/* Transformation matrices						    */
/*--------------------------------------------------------------------------*/

typedef struct {
   float a, b, c, d, e, f;
} Matrix;

/*-------------------------------------------------------------------------*/
/* Some convenience functions for matrix manipulation			   */
/*-------------------------------------------------------------------------*/

#define DCTM  &areastruct.MatStack[areastruct.MatStackDepth - 1]
#define DCTM2 &areastruct.MatStack[areastruct.MatStackDepth - 2]
#define UPopCTM() areastruct.MatStackDepth--

/*-------------------------------------------------------------------------*/
/* button tap/press definitions 					   */
/*-------------------------------------------------------------------------*/

#define PRESSTIME	150	/* milliseconds of push to be a "press" */

/*-------------------------------------------------------------------------*/
/* structures for managing the popup prompt widgets 			   */
/*-------------------------------------------------------------------------*/

typedef struct {
   Widget	button;
   int		foreground;
   void		(*buttoncall)();
   void		*dataptr;
} buttonsave;

typedef struct {
   Widget	popup;
   Widget	textw;
   void		(*setvalue)();
   buttonsave	*buttonptr;
   Boolean	fileprompt;
} popupstruct;

typedef struct {
   Widget	textw;
   Widget	buttonw;
   void		(*setvalue)();
   void		*dataptr;
} propstruct;

/*-------------------------------------------------------------------------*/
/* Initial Resource Management						   */
/*-------------------------------------------------------------------------*/

typedef struct {
   /* color scheme 1 */
   Pixel	fg, bg;
   Pixel	gridpix, snappix, selectpix, axespix;
   Pixel	buttonpix, querypix;
   Pixel	auxpix, barpix, bboxpix;

   /* color scheme 2 */

   Pixel	fg2, bg2;
   Pixel	gridpix2, snappix2, selectpix2, axespix2;
   Pixel	buttonpix2, querypix2;
   Pixel	auxpix2, barpix2, bboxpix2;

   int		width, height;
   XFontStruct	*xcfont, *helpfont, *filefont, *textfont, *titlefont;
} ApplicationData, *ApplicationDataPtr;

/*-------------------------------------------------------------------------*/
/* Macros for GC color and function handling				   */
/*-------------------------------------------------------------------------*/

#define XTopSetForeground(a) if (a == DEFAULTCOLOR) XSetForeground(dpy, \
	areastruct.gc, FOREGROUND); else XSetForeground(dpy, areastruct.gc, a)

#define XSetXORFg(a,b) if (a == DEFAULTCOLOR) XSetForeground(dpy, \
	areastruct.gc, FOREGROUND ^ b); else XSetForeground(dpy, areastruct.gc,\
	a ^ b)

#define XcSetFunction(z) XSetFunction(dpy, areastruct.gc, z); \
	areastruct.gctype = z

#define XcSetForeground(z) XTopSetForeground(z); areastruct.gccolor = \
	((z) == DEFAULTCOLOR) ? FOREGROUND : (z)

#define XcSetXORFg(y,z) XSetXORFg(y,z); areastruct.gccolor = \
	((y) == DEFAULTCOLOR) ? (FOREGROUND ^ z) : (y ^ z)

/*----------------------------------------------------------------------*/
/* Structure for maintaining list of colors				*/
/*----------------------------------------------------------------------*/

typedef struct {
   Widget	cbutton;
   XColor	color;
} colorindex;

/*----------------------------------------------------------------------*/
/* Font information structure						*/
/*----------------------------------------------------------------------*/
/* Flags:   bit  description						*/
/* 	    0    bold = 1,  normal = 0					*/
/*	    1    italic = 1, normal = 0					*/
/*	    2    <reserved, possibly for narrow font type>		*/
/*	    3    drawn = 1,  PostScript = 0				*/
/* 	    4	 <reserved, possibly for LaTeX font type>		*/
/*	    5    special encoding = 1, Standard Encoding = 0		*/
/*	    6	 ISOLatin1 = 2, ISOLatin2 = 3				*/
/*	    7    <reserved for other encoding schemes>			*/
/*----------------------------------------------------------------------*/

typedef struct {
   char *psname;
   char *family;
   float scale;
   unsigned short flags;
   objectptr *encoding;
} fontinfo;

/*-------------------------------------------------------------------------*/
/* the main globally-accessible data structure				   */
/* This structure holds all the critical data needed by the drawing window */
/*-------------------------------------------------------------------------*/

typedef struct {

   /* widgets and X11 parameters */
   Widget	area;
   Widget	scrollbarh, scrollbarv;
   GC 		gc;
   int		gccolor, gctype;
   XtIntervalId time_id;

   /* global page parameters */
   short	width, height;
   short	page, pages;
   float	*vscale;	/* pointer to proper scale */
   XPoint	*lowerleft;

   /* per-drawing-page parameters */
   char		**filename;	/* file to save as */
   float	*wirewidth;
   float	*outscale;
   float	*viewscale;
   float	*gridspace;
   float	*snapspace;
   short	*orient;
   short	*pmode;
   short	*coordstyle;
   XPoint	*drawingscale;
   XPoint	*pcorner;
   XPoint	*pagesize;	/* size of page to print on */

   /* global option defaults */
   float	textscale;
   short	psfont;
   short	justify;
   short	style;
   int		color;
   Boolean	manhatn;
   Boolean	boxedit;
   Boolean	snapto;
   Boolean	bboxon;
   Boolean	center;
   Boolean	gridon;
   Boolean	axeson;
   Boolean	invert;
#ifdef SCHEMA
   Boolean	schemon;
#endif

   /* buffers, pages, and associated variables */
   XPoint	save, origin;
   short	selects;
   short	*selectlist;
   objinstptr	topobject;
   objinstptr	*hierarchy;
   short	hlevels;
   short	builtins, userobjs, deletes, fontobjs;
   objectptr	*library, *userlib, *delbuffer, *fontlib;
   objectptr	editstack, *pagelist, libtop[LIBS];
   float	tempvscale, libvscale;
   XPoint	temppcorner, libpcorner;
   Matrix	*MatStack;
   short	MatStackDepth;
   short	editcycle;
   short	editpart, editsubpart;
} Clientdata, *Clientptr;

/*-------------------------------------------------------------------------*/
